import signal
import threading
import time
from abc import ABC, abstractmethod

from kaq_quant_common.utils.logger_utils import get_logger


# 通用的抽象类，包装一层ws操作
class WsWrapper(ABC):
    def __init__(self):
        # 初始化
        self.init()
        # 用来标记最后一次收到数据的时间，超过指定时间没有收到数据，认为连接断开
        self._keep_alive_deadline = 0

    # 初始化
    def init(self):
        # 执行初始化
        self._do_init()

    @abstractmethod
    def _do_init(self):
        pass

    # 开始，需要确保这个方法在op线程执行，执行这个方法会阻塞当前线程！
    def start(self):
        # 全局退出事件，用于传递终止信号
        exit_event = threading.Event()
        self._exit_event = exit_event

        logger = get_logger(self)

        #
        def handle_terminate_signal(signum, frame=None):
            """信号处理函数：捕获终止信号并触发退出事件"""
            logger.info(f"收到终止信号 {signum}")
            exit_event.set()
            # 优雅地关闭
            try:
                self.close()
            except Exception as e:
                pass

        # 监听信号
        # SIGTERM：Dagster通常发送此信号进行终止
        # SIGINT：对应Ctrl+C，用于本地测试
        signal.signal(signal.SIGTERM, handle_terminate_signal)
        signal.signal(signal.SIGINT, handle_terminate_signal)

        self._do_start()

        # 监听退出事件
        while not exit_event.is_set():
            time.sleep(1)
            if self._check_exit():
                # 假装发送信号
                handle_terminate_signal('exit')
            if self._keep_alive_deadline > 0 and time.time() > self._keep_alive_deadline:
                logger.warning("WsWrapper 收据接收超时，需要停止 WsWrapper")
                # 假装发送信号
                handle_terminate_signal('recv time out')

        logger.warning("WsWrapper 线程退出")

    # 子类实现，判断连接是否断开，断开返回True
    def _check_exit(self):
        return False

    @abstractmethod
    def _do_start(self):
        pass

    # 断开连接，主动关闭
    def close(self):
        self._do_close()

    @abstractmethod
    def _do_close(self):
        pass

    def keep_alive(self, deadline: int = 60):
        """
        保持连接alive，超过指定时间没有收到数据，认为连接断开
        """
        cur_time = time.time()
        self._keep_alive_deadline = cur_time + deadline



import os
import dolphindb as ddb
from kaq_quant_common.utils import yml_utils
import pandas as pd
import threading
from kaq_quant_common.utils.logger_utils import get_logger
import traceback

mutex = threading.Lock()

class KaqQuantDdbStreamWriteRepository:
    '''
    定义 asof_join的级联方式, 合并数据到一起, 然后可以订阅判断
    '''
    def __init__(self, host, port, user, passwd):
        self.logger = get_logger(self)
        '''
        创建ddb连接 && 添加ddb流数据表支持
        '''
        try:
            mutex.acquire()
            self.session = ddb.session(enableASYNC=True)
            self.session.connect(host, port, user, passwd, tryReconnectNums=10, reconnect=True, keepAliveTime=1000, readTimeout=100, writeTimeout=5)
            self.session.enableStreaming(threadCount=5)
            # self.pool = ddb.DBConnectionPool(host, port, userid=user, password=passwd, loadBalance=True, reConnect=True, tryReconnectNums=5, sqlStd=SqlStd.MySQL)

            # 需要注意的是 fetchSize 取值不能小于 8192 （记录条数）
            self.size = 8192
        except Exception as e:
            self.logger.error(f'KaqQuantDdbStreamWriteRepository.__init__ is occured error: {str(e)} - {str(traceback.format_exc())}')
        finally:
            mutex.release()
        
    def save2stream(self, ddb_table_name: str, df : pd.DataFrame):
        '''
        调用此方法之前, 需要将dataframe中的字符串类型的值 ，添加引号
        '''
        # 遍历每列的数据类型
        for column, dtype in df.dtypes.items():
            if dtype == 'object' or dtype == 'str':
                df[column] = '\'' + df[column] + '\''
        for index, row in df.iterrows():
            script = f"insert into {ddb_table_name} values({', '.join(str(x) for x in row.values)})"
            try:
                self.session.run(script, clearMemory=True)
            except Exception as e:
                self.logger.error(f'KaqQuantDdbStreamWriteRepository.save2stream is occured error: tableName is {ddb_table_name} - {str(e)} - {str(traceback.format_exc())}')


if __name__ == '__main__':
    host, port, user, passwd = yml_utils.get_ddb_info(os.getcwd())
    kaq = KaqQuantDdbStreamWriteRepository(host, port, user, passwd)


import os

_path = os.getcwd()

'''
根据异常错误判断是否重试
'''


def retry_if_io_error(exception):
    print("【zhima_util】异常错误为", exception)
    return isinstance(exception, IOError) or isinstance(exception, OSError)

'''
这里只是为了提供一个重试的方式，不一定要执行代理方式
'''


def retry_if_env_error(exception):
    return isinstance(exception, EnvironmentError)





import sqlite3
import os
import threading
import pandas as pd


class SqliteRepository:

    def __init__(self):
        print('sqlite\'s db init ............')
        self.con = None
        self.cur = None
        self.thread_lock = threading.Lock()

    def connect(self, db_name):
        try:
            self.thread_lock.acquire(blocking=True)
            db_path = os.getcwd() + os.sep + 'kaq_binance_sqlite3_db'
            if not os.path.exists(db_path):
                os.mkdir(db_path)

            sqlite3_db = db_path + os.sep + db_name
            self.con = sqlite3.connect(sqlite3_db, check_same_thread=False)
        except Exception as e:
            print('【SqliteRepository.connect】异常', db_name, e)
        finally:
            # print('【SqliteRepository.connect】释放锁', db_name)
            self.thread_lock.release()

    def get_table_info(self, sql):
        '''
        提取表的字段
        '''
        try:
            self.thread_lock.acquire(True)
            self.cur = self.con.cursor()
            self.cur.execute(sql)
            person_all = self.cur.fetchall()
            # 返回数据
            return person_all
        except Exception as e:
            self.con.rollback()
            print('【sqllite查询失败】', sql, e)
        finally:
            # print('【SqliteRepository.select_data】释放锁', params)
            self.cur.close()
            self.thread_lock.release()

    def is_table_exists(self, table_name):
        '''
        表是否存在
        '''
        try:
            self.thread_lock.acquire(True)
            self.cur = self.con.cursor()
            self.cur.execute("PRAGMA table_info({})".format(table_name))
            person_all = self.cur.fetchone()
            # 返回数据
            if person_all is not None:
                return True
            else:
                return False
        except Exception as e:
            self.con.rollback()
            print('【sqllite查询失败】', table_name, e)
        finally:
            # print('【SqliteRepository.select_data】释放锁', params)
            self.cur.close()
            self.thread_lock.release()

    def close(self):
        try:
            self.thread_lock.acquire(blocking=True)
            self.con.close()
        except Exception as e:
            print('【SqliteRepository.close】异常', e)
        finally:
            # print('【SqliteRepository.close】释放锁')
            self.thread_lock.release()

    def create_table(self, create_table_sql):
        try:
            self.thread_lock.acquire(blocking=True)
            self.cur = self.con.cursor()
            self.cur.execute(create_table_sql)
            self.con.commit()
            # print('【创建表成功】', create_table_sql)
        except Exception as e:
            print('【sqllite创建表失败】', create_table_sql, e)
        finally:
            # print('【SqliteRepository.create_table】释放锁', create_table_sql)
            self.cur.close()
            self.thread_lock.release()

    def insert_tab(self, insert_sql, params):
        try:
            self.thread_lock.acquire(True)
            self.cur = self.con.cursor()
            self.cur.execute(insert_sql, params)
            self.con.commit()
            # print('插入成功')
        except Exception as e:
            self.con.rollback()
            print("【sqllite插入失败】", insert_sql, params, e)
        finally:
            # print('【SqliteRepository.insert_tab】释放锁', params)
            self.cur.close()
            self.thread_lock.release()

    def delete_data(self, delete_sql, params):
        try:
            self.thread_lock.acquire(True)
            self.cur = self.con.cursor()
            self.cur.execute(delete_sql, params)
            self.con.commit()
            # print('删除成功')
        except Exception as e:
            self.con.rollback()
            print('【sqllite删除失败】', e, delete_sql, params)
        finally:
            # print('【SqliteRepository.delete_data】释放锁', params)
            self.cur.close()
            self.thread_lock.release()

    def update_data(self, update_sql, params):
        try:
            self.thread_lock.acquire(True)
            self.cur = self.con.cursor()
            self.cur.execute(update_sql, params)
            self.con.commit()
            # print('修改成功')
        except Exception as e:
            self.con.rollback()
            # print('【sqllite更新失败】', e, update_sql, params)
        finally:
            # print('【SqliteRepository.update_data】释放锁', params)
            self.cur.close()
            self.thread_lock.release()

    def select_data(self, select_sql, params):
        try:
            self.thread_lock.acquire(True)
            self.cur = self.con.cursor()
            self.cur.execute(select_sql, params)
            person_all = self.cur.fetchall()
            # 返回数据
            return person_all
        except Exception as e:
            self.con.rollback()
            print('【sqllite查询失败】', select_sql, params, e)
        finally:
            # print('【SqliteRepository.select_data】释放锁', params)
            self.cur.close()
            self.thread_lock.release()

    # 将pdFrame转为table
    def pd_transfer_table(self, pd_df, table_name):
        try:
            self.thread_lock.acquire(True)
            pd.io.sql.to_sql(pd_df, table_name, self.con, if_exists='replace')
        except Exception as e:
            print('将pdFrame转为table失败', e)
        finally:
            self.thread_lock.release()


sqliteRepository = SqliteRepository()





import os
import dolphindb as ddb
from kaq_quant_common.utils import yml_utils
import pandas as pd
import threading
from kaq_quant_common.utils.logger_utils import get_logger
import traceback

mutex = threading.Lock()

class KaqQuantDdbStreamReadRepository:
    '''
    定义 asof_join的级联方式, 合并数据到一起, 然后可以订阅判断
    '''
    def __init__(self, host, port, user, passwd):
        self.logger = get_logger(self)
        '''
        创建ddb连接 && 添加ddb流数据表支持
        '''
        try:
            mutex.acquire()
            # 程序读取的时候 同步读取即可, ddb的异步读取，可以使用订阅的方式来实现
            self.session = ddb.session(enableASYNC=False)
            self.session.connect(host, port, user, passwd)
            self.session.enableStreaming()

            # 需要注意的是 fetchSize 取值不能小于 8192 （记录条数）
            self.size = 8192
        except Exception as e:
            self.logger.error(f'KaqQuantDdbStreamReadRepository.__init__ is occured error: {str(e)} - {str(traceback.format_exc())}')
        finally:
            mutex.release()

    def count(self, query: str) -> int:
        '''
        获取数量, 再去获取总数据
        sql语句中必须返回的是 count 字段
        '''
        try:
            number_blocks = self.session.run(query, fetchSize=self.size)
            number = 0
            while number_blocks.hasNext():
                number = number_blocks.read()['count'][0]
            return number
        except Exception as e:
            self.logger.error(f'KaqQuantDdbStreamReadRepository.countStream is occured error: {str(e)} - {str(traceback.format_exc())}')
        return -1

    def cal_page(self, number: int):
        page = number / self.size
        if number % self.size > 0:
            page = page + 1
        return int(page)

    def query(self, query: str) -> pd.DataFrame:
        '''
        从流数据表中获取数据
        '''
        try:
            block = self.session.run(query, fetchSize=self.size)
            if block is None:
                return pd.DataFrame()
            big_df = pd.DataFrame()
            while block.hasNext():
                temp_df = block.read()
                big_df = pd.concat([big_df, temp_df])
            big_df = big_df.reset_index()
            return big_df
        except Exception as e:
            self.logger.error(f'KaqQuantDdbStreamReadRepository.getStreamQuery is occured error: {str(e)} - {str(traceback.format_exc())}')
        return pd.DataFrame()


if __name__ == '__main__':
    host, port, user, passwd = yml_utils.get_ddb_info(os.getcwd())
    kaq = KaqQuantDdbStreamReadRepository(host, port, user, passwd)
    # kaq.drop_streaming()
    count = kaq.countStream(f"select count(*) from kaq_binance_force_order_streaming where symbol == 'FIOUSDT'")
    df = kaq.getStreamQuery(query=f"select * from kaq_binance_force_order_streaming where symbol == 'FIOUSDT' order by create_time asc")
    print(df.head(5))


import json
import os
from typing import Any, List
import pandas as pd
import redis
import threading
from kaq_quant_common.utils import yml_utils
from loguru import logger
import traceback

mutex = threading.Lock()

class KaqQuantRedisRepository:
    '''
    定义 redis操作
    '''
    def __init__(self, host='localhost', port=6379, password=None, db=0):
        '''
        redis连接池
        '''
        try:
            mutex.acquire()
            # 创建连接池
            pool = redis.ConnectionPool(host=host, port=port, db=db, password=password, max_connections=3, decode_responses=True)

            # 共享的 Redis 客户端
            self.client = redis.StrictRedis(connection_pool=pool)
        except Exception as e:
            logger.error(f"【创建redis连接】错误异常: {str(e)} - {str(traceback.format_exc())}")
        finally:
            mutex.release()
            
    # 显式代理 Redis 常用方法
    def set(self, name: str, value: Any, ex: int = None, px: int = None, nx: bool = False, xx: bool = False):
        return self.client.set(name, value, ex, px, nx, xx)

    def get(self, name: str) -> Any:
        return self.client.get(name)

    def delete(self, *names: str) -> int:
        return self.client.delete(*names)

    def keys(self, pattern: str) -> list:
        return self.client.keys(pattern)

    def exists(self, name: str) -> bool:
        return self.client.exists(name)
    
    def lrange(self, name: str) -> pd.DataFrame:
        result = self.client.lrange(name, 0, -1)
        return pd.DataFrame([json.loads(item) for item in result]) if result else pd.DataFrame()
    
    def rpush(self, name: str, df: pd.DataFrame) -> pd.DataFrame:
        # 每次都先删除，后添加
        self.client.delete(name)
        json_data = [json.dumps(row.to_dict()) for _, row in df.iterrows()]
        self.client.rpush(name, *json_data)
        return 
    
    def get_keys_by_pattern(self, pattern: str) -> List[str]:
        '''
        根据模糊前缀获取 Redis 中匹配的键
        :param pattern: 模糊匹配的模式，支持 * 通配符
        :return: 匹配的键列表
        '''
        try:
            # 获取匹配的键
            keys = self.client.keys(pattern)
            return keys
        except Exception as e:
            logger.error(f"获取匹配键时出错: {str(e)} - {str(traceback.format_exc())}")
            return []

    def get_values_by_pattern(self, pattern: str) -> pd.DataFrame:
        '''
        批量获取 Redis 中符合模糊前缀匹配的键对应的 JSON 数据，并构建 pandas DataFrame
        :param pattern: 模糊匹配的模式，支持 * 通配符
        :return: 返回构建的 pandas DataFrame
        '''
        try:
            # 获取所有匹配的键
            keys = self.get_keys_by_pattern(pattern)
            
            # 批量获取对应的 JSON 数据
            json_data_list = self.client.mget(keys)
            
            # 将 JSON 数据转换为字典列表
            data_list = [json.loads(data.replace("'", '"')) if data else None for data in json_data_list]
            
            # 去除为 None 的数据
            data_list = [data for data in data_list if data is not None]

            # 构建 pandas DataFrame
            df = pd.DataFrame(data_list)
            return df
        except Exception as e:
            logger.error(f"获取并转换 JSON 数据时出错: {str(e)} - {str(traceback.format_exc())}")
            return pd.DataFrame()  # 返回空 DataFrame
            
    def __getattr__(self, name):
        '''
        使得对象在没有明确调用时，能够直接调用 Redis 的方法
        '''
        return getattr(self.client, name)

if __name__ == '__main__':
    host, port, passwd = yml_utils.get_redis_info(os.getcwd())
    KaqQuantRedisRepository = KaqQuantRedisRepository(host=host, port=port, password=passwd)
    # value = KaqQuantRedisRepository.get('test')
    df = KaqQuantRedisRepository.get_values_by_pattern('kaq_binance_commision_rate_*')
    print(df)


# MyTT 麦语言-通达信-同花顺指标实现     https://github.com/mpquant/MyTT
# MyTT高级函数验证版本：               https://github.com/mpquant/MyTT/blob/main/MyTT_plus.py
# Python2老版本pandas特别的MyTT：      https://github.com/mpquant/MyTT/blob/main/MyTT_python2.py 
# V2.1  2021-6-6   新增 BARSLAST函数 SLOPE,FORCAST线性回归预测函数
# V2.3  2021-6-13  新增 TRIX,DPO,BRAR,DMA,MTM,MASS,ROC,VR,ASI等指标
# V2.4  2021-6-27  新增 EXPMA,OBV,MFI指标, 改进SMA核心函数(核心函数彻底无循环)
# V2.7  2021-11-21 修正 SLOPE,BARSLAST,函数,新加FILTER,LONGCROSS, 感谢qzhjiang对SLOPE,SMA等函数的指正
# V2.8  2021-11-23 修正 FORCAST,WMA函数,欢迎qzhjiang,stanene,bcq加入社群，一起来完善myTT库
# V2.9  2021-11-29 新增 HHVBARS,LLVBARS,CONST, VALUEWHEN功能函数
# V2.92 2021-11-30 新增 BARSSINCEN函数,现在可以 pip install MyTT 完成安装   
# V3.0  2021-12-04 改进 DMA函数支持序列,新增XS2 薛斯通道II指标
# V3.1  2021-12-19 新增 TOPRANGE,LOWRANGE一级函数

# 以下所有函数如无特别说明，输入参数S均为numpy序列或者列表list，N为整型int
# 应用层1级函数完美兼容通达信或同花顺，具体使用方法请参考通达信
import numpy as np
import pandas as pd


# ------------------ 0级：核心工具函数 --------------------------------------------
def RD(N, D=3):   return np.round(N, D)  # 四舍五入取3位小数


def RET(S, N=1):  return np.array(S)[-N]  # 返回序列倒数第N个值,默认返回最后一个


def ABS(S):      return np.abs(S)  # 返回N的绝对值


def LN(S):       return np.log(S)  # 求底是e的自然对数,


def POW(S, N):    return np.power(S, N)  # 求S的N次方


def SQRT(S):     return np.sqrt(S)  # 求S的平方根


def MAX(S1, S2):  return np.maximum(S1, S2)  # 序列max


def MIN(S1, S2):  return np.minimum(S1, S2)  # 序列min


def IF(S, A, B):   return np.where(S, A, B)  # 序列布尔判断 return=A  if S==True  else  B


def REF(S, N=1):  # 对序列整体下移动N,返回序列(shift后会产生NAN)
    return pd.Series(S).shift(N).values


def DIFF(S, N=1):  # 前一个值减后一个值,前面会产生nan
    return pd.Series(S).diff(N).values  # np.diff(S)直接删除nan，会少一行


def STD(S, N):  # 求序列的N日标准差，返回序列
    return pd.Series(S).rolling(N).std(ddof=0).values


def SUM(S, N):  # 对序列求N天累计和，返回序列    N=0对序列所有依次求和
    return pd.Series(S).rolling(N).sum().values if N > 0 else pd.Series(S).cumsum().values


def CONST(S):  # 返回序列S最后的值组成常量序列
    return np.full(len(S), S[-1])


def HHV(S, N):  # HHV(C, 5) 最近5天收盘最高价
    return pd.Series(S).rolling(N).max().values


def LLV(S, N):  # LLV(C, 5) 最近5天收盘最低价
    return pd.Series(S).rolling(N).min().values


def HHVBARS(S, N):  # 求N周期内S最高值到当前周期数, 返回序列
    return pd.Series(S).rolling(N).apply(lambda x: np.argmax(x[::-1]), raw=True).values


def LLVBARS(S, N):  # 求N周期内S最低值到当前周期数, 返回序列
    return pd.Series(S).rolling(N).apply(lambda x: np.argmin(x[::-1]), raw=True).values


def MA(S, N):  # 求序列的N日简单移动平均值，返回序列
    return pd.Series(S).rolling(N).mean().values


def EMA(S, N):  # 指数移动平均,为了精度 S>4*N  EMA至少需要120周期     alpha=2/(span+1)
    return pd.Series(S).ewm(span=N, adjust=False).mean().values


def SMA(S, N, M=1):  # 中国式的SMA,至少需要120周期才精确 (雪球180周期)    alpha=1/(1+com)
    return pd.Series(S).ewm(alpha=M / N, adjust=False).mean().values  # com=N-M/M


def WMA(S, N):  # 通达信S序列的N日加权移动平均 Yn = (1*X1+2*X2+3*X3+...+n*Xn)/(1+2+3+...+n)
    return pd.Series(S).rolling(N).apply(lambda x: x[::-1].cumsum().sum() * 2 / N / (N + 1), raw=True).values


def DMA(S, A):  # 求S的动态移动平均，A作平滑因子,必须 0<A<1  (此为核心函数，非指标）
    if isinstance(A, (int, float)): return pd.Series(S).ewm(alpha=A, adjust=False).mean().values
    A = np.array(A);
    A[np.isnan(A)] = 1.0;
    Y = np.zeros(len(S));
    Y[0] = S[0]
    for i in range(1, len(S)): Y[i] = A[i] * S[i] + (1 - A[i]) * Y[i - 1]  # A支持序列 by jqz1226
    return Y


def AVEDEV(S, N):  # 平均绝对偏差  (序列与其平均值的绝对差的平均值)
    return pd.Series(S).rolling(N).apply(lambda x: (np.abs(x - x.mean())).mean()).values


def SLOPE(S, N):  # 返S序列N周期回线性回归斜率
    return pd.Series(S).rolling(N).apply(lambda x: np.polyfit(range(N), x, deg=1)[0], raw=True).values


def FORCAST(S, N):  # 返回S序列N周期回线性回归后的预测值， jqz1226改进成序列出
    return pd.Series(S).rolling(N).apply(lambda x: np.polyval(np.polyfit(range(N), x, deg=1), N - 1), raw=True).values


def LAST(S, A, B):  # 从前A日到前B日一直满足S_BOOL条件, 要求A>B & A>0 & B>=0
    return np.array(pd.Series(S).rolling(A + 1).apply(lambda x: np.all(x[::-1][B:]), raw=True), dtype=bool)


# ------------------   1级：应用层函数(通过0级核心函数实现）使用方法请参考通达信--------------------------------
def COUNT(S, N):  # COUNT(CLOSE>O, N):  最近N天满足S_BOO的天数  True的天数
    return SUM(S, N)


def EVERY(S, N):  # EVERY(CLOSE>O, 5)   最近N天是否都是True
    return IF(SUM(S, N) == N, True, False)


def EXIST(S, N):  # EXIST(CLOSE>3010, N=5)  n日内是否存在一天大于3000点
    return IF(SUM(S, N) > 0, True, False)


def FILTER(S, N):  # FILTER函数，S满足条件后，将其后N周期内的数据置为0, FILTER(C==H,5)
    for i in range(len(S)): S[i + 1:i + 1 + N] = 0 if S[i] else S[i + 1:i + 1 + N]
    return S  # 例：FILTER(C==H,5) 涨停后，后5天不再发出信号


def BARSLAST(S):  # 上一次条件成立到当前的周期, BARSLAST(C/REF(C,1)>=1.1) 上一次涨停到今天的天数
    M = np.concatenate(([0], np.where(S, 1, 0)))
    for i in range(1, len(M)):  M[i] = 0 if M[i] else M[i - 1] + 1
    return M[1:]


def BARSLASTCOUNT(S):  # 统计连续满足S条件的周期数        by jqz1226
    rt = np.zeros(len(S) + 1)  # BARSLASTCOUNT(CLOSE>OPEN)表示统计连续收阳的周期数
    for i in range(len(S)): rt[i + 1] = rt[i] + 1 if S[i] else rt[i + 1]
    return rt[1:]


def BARSSINCEN(S, N):  # N周期内第一次S条件成立到现在的周期数,N为常量  by jqz1226
    return pd.Series(S).rolling(N).apply(lambda x: N - 1 - np.argmax(x) if np.argmax(x) or x[0] else 0,
                                         raw=True).fillna(0).values.astype(int)


def CROSS(S1, S2):  # 判断向上金叉穿越 CROSS(MA(C,5),MA(C,10))  判断向下死叉穿越 CROSS(MA(C,10),MA(C,5))
    return np.concatenate(([False], np.logical_not((S1 > S2)[:-1]) & (S1 > S2)[1:]))  # 不使用0级函数,移植方便  by jqz1226


def LONGCROSS(S1, S2, N):  # 两条线维持一定周期后交叉,S1在N周期内都小于S2,本周期从S1下方向上穿过S2时返回1,否则返回0
    return np.array(np.logical_and(LAST(S1 < S2, N, 1), (S1 > S2)), dtype=bool)  # N=1时等同于CROSS(S1, S2)


def VALUEWHEN(S, X):  # 当S条件成立时,取X的当前值,否则取VALUEWHEN的上个成立时的X值   by jqz1226
    return pd.Series(np.where(S, X, np.nan)).ffill().values


def BETWEEN(S, A, B):  # S处于A和B之间时为真。 包括 A<S<B 或 A>S>B
    return ((A < S) & (S < B)) | ((A > S) & (S > B))


def TOPRANGE(S):  # TOPRANGE(HIGH)表示当前最高价是近多少周期内最高价的最大值 by jqz1226
    rt = np.zeros(len(S))
    for i in range(1, len(S)):  rt[i] = np.argmin(np.flipud(S[:i] < S[i]))
    return rt.astype('int')


def LOWRANGE(S):  # LOWRANGE(LOW)表示当前最低价是近多少周期内最低价的最小值 by jqz1226
    rt = np.zeros(len(S))
    for i in range(1, len(S)):  rt[i] = np.argmin(np.flipud(S[:i] > S[i]))
    return rt.astype('int')


# ------------------   2级：技术指标函数(全部通过0级，1级函数实现） ------------------------------
def MACD(CLOSE, SHORT=12, LONG=26, M=9):  # EMA的关系，S取120日，和雪球小数点2位相同
    DIF = EMA(CLOSE, SHORT) - EMA(CLOSE, LONG);
    DEA = EMA(DIF, M);
    MACD = (DIF - DEA) * 2
    return RD(DIF), RD(DEA), RD(MACD)


def KDJ(CLOSE, HIGH, LOW, N=9, M1=3, M2=3):  # KDJ指标
    RSV = (CLOSE - LLV(LOW, N)) / (HHV(HIGH, N) - LLV(LOW, N)) * 100
    K = EMA(RSV, (M1 * 2 - 1));
    D = EMA(K, (M2 * 2 - 1));
    J = K * 3 - D * 2
    return K, D, J


def RSI(CLOSE, N=24):  # RSI指标,和通达信小数点2位相同
    DIF = CLOSE - REF(CLOSE, 1)
    return RD(SMA(MAX(DIF, 0), N) / SMA(ABS(DIF), N) * 100)


def WR(CLOSE, HIGH, LOW, N=10, N1=6):  # W&R 威廉指标
    WR = (HHV(HIGH, N) - CLOSE) / (HHV(HIGH, N) - LLV(LOW, N)) * 100
    WR1 = (HHV(HIGH, N1) - CLOSE) / (HHV(HIGH, N1) - LLV(LOW, N1)) * 100
    return RD(WR), RD(WR1)


def BIAS(CLOSE, L1=6, L2=12, L3=24):  # BIAS乖离率
    BIAS1 = (CLOSE - MA(CLOSE, L1)) / MA(CLOSE, L1) * 100
    BIAS2 = (CLOSE - MA(CLOSE, L2)) / MA(CLOSE, L2) * 100
    BIAS3 = (CLOSE - MA(CLOSE, L3)) / MA(CLOSE, L3) * 100
    return RD(BIAS1), RD(BIAS2), RD(BIAS3)


def BOLL(CLOSE, N=20, P=2):  # BOLL指标，布林带
    MID = MA(CLOSE, N);
    UPPER = MID + STD(CLOSE, N) * P
    LOWER = MID - STD(CLOSE, N) * P
    return RD(UPPER), RD(MID), RD(LOWER)


def PSY(CLOSE, N=12, M=6):
    PSY = COUNT(CLOSE > REF(CLOSE, 1), N) / N * 100
    PSYMA = MA(PSY, M)
    return RD(PSY), RD(PSYMA)


def CCI(CLOSE, HIGH, LOW, N=14):
    TP = (HIGH + LOW + CLOSE) / 3
    return (TP - MA(TP, N)) / (0.015 * AVEDEV(TP, N))


def ATR(CLOSE, HIGH, LOW, N=20):  # 真实波动N日平均值
    TR = MAX(MAX((HIGH - LOW), ABS(REF(CLOSE, 1) - HIGH)), ABS(REF(CLOSE, 1) - LOW))
    return MA(TR, N)


def BBI(CLOSE, M1=3, M2=6, M3=12, M4=20):  # BBI多空指标
    return (MA(CLOSE, M1) + MA(CLOSE, M2) + MA(CLOSE, M3) + MA(CLOSE, M4)) / 4


def DMI(CLOSE, HIGH, LOW, M1=14, M2=6):  # 动向指标：结果和同花顺，通达信完全一致
    TR = SUM(MAX(MAX(HIGH - LOW, ABS(HIGH - REF(CLOSE, 1))), ABS(LOW - REF(CLOSE, 1))), M1)
    HD = HIGH - REF(HIGH, 1);
    LD = REF(LOW, 1) - LOW
    DMP = SUM(IF((HD > 0) & (HD > LD), HD, 0), M1)
    DMM = SUM(IF((LD > 0) & (LD > HD), LD, 0), M1)
    PDI = DMP * 100 / TR;
    MDI = DMM * 100 / TR
    ADX = MA(ABS(MDI - PDI) / (PDI + MDI) * 100, M2)
    ADXR = (ADX + REF(ADX, M2)) / 2
    return PDI, MDI, ADX, ADXR


def TAQ(HIGH, LOW, N):  # 唐安奇通道(海龟)交易指标，大道至简，能穿越牛熊
    UP = HHV(HIGH, N);
    DOWN = LLV(LOW, N);
    MID = (UP + DOWN) / 2
    return UP, MID, DOWN


def KTN(CLOSE, HIGH, LOW, N=20, M=10):  # 肯特纳交易通道, N选20日，ATR选10日
    MID = EMA((HIGH + LOW + CLOSE) / 3, N)
    ATRN = ATR(CLOSE, HIGH, LOW, M)
    UPPER = MID + 2 * ATRN;
    LOWER = MID - 2 * ATRN
    return UPPER, MID, LOWER


def TRIX(CLOSE, M1=12, M2=20):  # 三重指数平滑平均线
    TR = EMA(EMA(EMA(CLOSE, M1), M1), M1)
    TRIX = (TR - REF(TR, 1)) / REF(TR, 1) * 100
    TRMA = MA(TRIX, M2)
    return TRIX, TRMA


def VR(CLOSE, VOL, M1=26):  # VR容量比率
    LC = REF(CLOSE, 1)
    return SUM(IF(CLOSE > LC, VOL, 0), M1) / SUM(IF(CLOSE <= LC, VOL, 0), M1) * 100


def EMV(HIGH, LOW, VOL, N=14, M=9):  # 简易波动指标
    VOLUME = MA(VOL, N) / VOL;
    MID = 100 * (HIGH + LOW - REF(HIGH + LOW, 1)) / (HIGH + LOW)
    EMV = MA(MID * VOLUME * (HIGH - LOW) / MA(HIGH - LOW, N), N);
    MAEMV = MA(EMV, M)
    return EMV, MAEMV


def DPO(CLOSE, M1=20, M2=10, M3=6):  # 区间震荡线
    DPO = CLOSE - REF(MA(CLOSE, M1), M2);
    MADPO = MA(DPO, M3)
    return DPO, MADPO


def BRAR(OPEN, CLOSE, HIGH, LOW, M1=26):  # BRAR-ARBR 情绪指标
    AR = SUM(HIGH - OPEN, M1) / SUM(OPEN - LOW, M1) * 100
    BR = SUM(MAX(0, HIGH - REF(CLOSE, 1)), M1) / SUM(MAX(0, REF(CLOSE, 1) - LOW), M1) * 100
    return AR, BR


def DFMA(CLOSE, N1=10, N2=50, M=10):  # 平行线差指标
    DIF = MA(CLOSE, N1) - MA(CLOSE, N2);
    DIFMA = MA(DIF, M)  # 通达信指标叫DMA 同花顺叫新DMA
    return DIF, DIFMA


def MTM(CLOSE, N=12, M=6):  # 动量指标
    MTM = CLOSE - REF(CLOSE, N);
    MTMMA = MA(MTM, M)
    return MTM, MTMMA


def MASS(HIGH, LOW, N1=9, N2=25, M=6):  # 梅斯线
    MASS = SUM(MA(HIGH - LOW, N1) / MA(MA(HIGH - LOW, N1), N1), N2)
    MA_MASS = MA(MASS, M)
    return MASS, MA_MASS


def ROC(CLOSE, N=12, M=6):  # 变动率指标
    ROC = 100 * (CLOSE - REF(CLOSE, N)) / REF(CLOSE, N);
    MAROC = MA(ROC, M)
    return ROC, MAROC


def EXPMA(CLOSE, N1=12, N2=50):  # EMA指数平均数指标
    return EMA(CLOSE, N1), EMA(CLOSE, N2);


def OBV(CLOSE, VOL):  # 能量潮指标
    return SUM(IF(CLOSE > REF(CLOSE, 1), VOL, IF(CLOSE < REF(CLOSE, 1), -VOL, 0)), 0) / 10000


def MFI(CLOSE, HIGH, LOW, VOL, N=14):  # MFI指标是成交量的RSI指标
    TYP = (HIGH + LOW + CLOSE) / 3
    V1 = SUM(IF(TYP > REF(TYP, 1), TYP * VOL, 0), N) / SUM(IF(TYP < REF(TYP, 1), TYP * VOL, 0), N)
    return 100 - (100 / (1 + V1))


def ASI(OPEN, CLOSE, HIGH, LOW, M1=26, M2=10):  # 振动升降指标
    LC = REF(CLOSE, 1);
    AA = ABS(HIGH - LC);
    BB = ABS(LOW - LC);
    CC = ABS(HIGH - REF(LOW, 1));
    DD = ABS(LC - REF(OPEN, 1));
    R = IF((AA > BB) & (AA > CC), AA + BB / 2 + DD / 4, IF((BB > CC) & (BB > AA), BB + AA / 2 + DD / 4, CC + DD / 4));
    X = (CLOSE - LC + (CLOSE - OPEN) / 2 + LC - REF(OPEN, 1));
    SI = 16 * X / R * MAX(AA, BB);
    ASI = SUM(SI, M1);
    ASIT = MA(ASI, M2);
    return ASI, ASIT


def XSII(CLOSE, HIGH, LOW, N=102, M=7):  # 薛斯通道II
    AA = MA((2 * CLOSE + HIGH + LOW) / 4, 5)  # 最新版DMA才支持 2021-12-4
    TD1 = AA * N / 100;
    TD2 = AA * (200 - N) / 100
    CC = ABS((2 * CLOSE + HIGH + LOW) / 4 - MA(CLOSE, 20)) / MA(CLOSE, 20)
    DD = DMA(CLOSE, CC);
    TD3 = (1 + M / 100) * DD;
    TD4 = (1 - M / 100) * DD
    return TD1, TD2, TD3, TD4

    # 望大家能提交更多指标和函数  https://github.com/mpquant/MyTT



import os
from typing import Union
import yaml
import pkgutil
import importlib.resources as resources
from loguru import logger


def read_config(pkg_name='kaq_quant_common'):
    config_path = pkgutil.get_data(pkg_name, f'config{os.sep}config.yaml')
    data = yaml.load(config_path, Loader=yaml.FullLoader)
    return data

def get(pkg_name='kaq_quant_common', *args):
    '''
    根据key获取配置，支持:
    - 单个值: get_by_key('mysql', 'host') → 获取 kaq.mysql.host
    - 多个值: get_by_key('mysql', ['host', 'port']) → 获取 (kaq.mysql.host, kaq.mysql.port)
    - 多层级: get_by_key('a', 'b', 'c') → 获取 kaq.a.b.c
    - 顶层多个值: get_by_key(['api_key', 'api_secret']) → 获取 (kaq.api_key, kaq.api_secret)
    '''
    if not args:  # 没有提供key参数
        return None
    data = read_config(pkg_name)
    if data is None or 'kaq' not in data:
        return None
    data = data['kaq']  # 已定位到kaq节点
    
    last_arg = args[-1]
    if isinstance(last_arg, list):
        # 最后一个参数是列表，获取多个key的值
        keys = last_arg
        path_parts = args[:-1]  # 路径部分
        if not keys:  # 列表为空
            return None
        
        # 遍历路径部分
        current_data = data
        for part in path_parts:
            if not isinstance(current_data, dict) or part not in current_data:
                return None
            current_data = current_data[part]
        
        # 检查并获取所有key的值
        for key in keys:
            if key not in current_data:
                return None
        return tuple(current_data[key] for key in keys)
    else:
        # 所有参数均为路径，获取单个值
        current_data = data
        for part in args:
            if not isinstance(current_data, dict) or part not in current_data:
                return None
            current_data = current_data[part]
        return current_data

def get_spot_list(pkg_name='kaq_quant_common'):
    '''
    获取合约对应的现货列表
    '''
    return get(pkg_name, 'spot_list')

def get_future_list(pkg_name='kaq_quant_common'):
    '''
    获取合约的symbol列表
    '''
    return get(pkg_name, 'future_list')

def get_api_key_secret(pkg_name='kaq_quant_common'):
    '''
    获取api_key,api_secret
    '''
    return get(pkg_name, ['api_key','api_secret']) 

def get_proxies(pkg_name='kaq_quant_common'):
    '''
    获取代理配置
    '''
    return get(pkg_name, 'proxies')

def get_mysql_info(pkg_name='kaq_quant_common'):
    '''
    mysql配置
    '''
    return get(pkg_name, 'mysql', ['host', 'port', 'user', 'passwd', 'database', 'charset'])

def get_redis_info(pkg_name='kaq_quant_common'):
    '''
    redis配置
    '''
    return get(pkg_name, 'redis', ['host', 'port', 'passwd'])

def get_posgresql_info(pkg_name='kaq_quant_common'):
    '''
    posgresql配置
    '''
    return get(pkg_name, 'posgresql', ['host', 'port', 'user', 'passwd', 'database', 'charset'])

def get_mysql_table_prefix(pkg_name='kaq_quant_common'):
    return get(pkg_name, 'mysql_table_prefix')

def get_ddb_info(pkg_name='kaq_quant_common'):
    '''
    ddb配置
    '''
    return get(pkg_name, 'ddb', ['host', 'port', 'user', 'passwd'])


def get_pkg_file(pkg=None, file_name=''):
    if pkg is None:
        logger.error('yml_utils.get_pkg_script pkg is None')
        return None
    '''
    读取文件
    '''
    with resources.path(pkg, file_name) as file_path:
        # file_path 是一个 Path 对象, 输出文件的绝对路径
        config_path = str(file_path)
    if os.path.exists(config_path):
        return config_path
    logger.error(f'yml_utils.get_pkg_script {config_path} is not exits!')
    return None
 

 
if __name__ == '__main__':
    kv = get_ddb_info()
    print(kv)




from enum import Enum

class KaqCommsionRateRedisPrefixEnum(Enum):
    binance_future = 'kaq_binance_futures_commission_rate'
    bybit_future = 'kaq_bybit_futures_commission_rate'
    okx_future = 'kaq_okx_futures_commission_rate'
    bitget_future = 'kaq_bitget_futures_commission_rate'
    gate_future = 'kaq_gate_futures_commission_rate'
    htx_future = 'kaq_gate_futures_commission_rate'
    
    binance_spot = 'kaq_binance_spot_commsion_rate'
    bybit_spot = 'kaq_bybit_spot_commsion_rate'
    okx_spot = 'kaq_okx_spot_commsion_rate'
    bitget_spot = 'kaq_bitget_spot_commsion_rate'
    gate_spot = 'kaq_gate_spot_commsion_rate'
    htx_spot = 'kaq_htx_spot_commsion_rate'
    
class KaqSpotInterestRateRedisPrefixEnum(Enum):
    
    binance_spot = 'kaq_binance_spot_interest_rate'
    bybit_spot = 'kaq_bybit_spot_interest_rate'
    okx_spot = 'kaq_okx_spot_interest_rate'
    bitget_spot = 'kaq_bitget_spot_interest_rate'
    gate_spot = 'kaq_gate_spot_interest_rate'
    htx_spot = 'kaq_htx_spot_interest_rate'

class KaqCoinDataEnum(Enum):
    '''
    枚举检测
    '''
    klines = 'klines' # klines
    global_long_short_account_ratio = 'global_long_short_account_ratio' # 多空持仓人数比
    open_interest_hist = 'open_interest_hist' # 合约持仓量历史
    taker_long_short_ratio = 'taker_long_short_ratio' # 合约主动买卖量
    top_long_short_account_ratio = 'top_long_short_account_ratio' # 大户账户数多空比
    top_long_short_position_ratio = 'top_long_short_position_ratio' # 大户持仓量多空比


import os
from kaq_quant_common.utils import yml_utils
import pandas as pd
import threading
import psycopg2
from kaq_quant_common.utils.logger_utils import get_logger

mutex = threading.Lock()

class KaqQuantPostgreSqlRepository:
    '''
    timescaleDB的操作类
    '''
    def __init__(self, host, port, user, passwd, database, charset='utf8'):
        self.logger = get_logger(self)
        # 数据库连接参数
        conn_params = {
            "dbname": database,
            "user": user,
            "password": passwd,
            "host": host,
            "port": port
        }
        # 创建写入引擎数据库连接池
        # 建立连接和游标
        self.conn = psycopg2.connect(**conn_params)
        self.cur = self.conn.cursor()
        
        # 关闭游标和连接
        # self.cur.close()
        # self.conn.close()

    def fetch_data(self, query):
        '''
        查询操作，输入查询语句
        '''
        try:
            # 执行查询
            self.cur.execute(query)
            rows = self.cur.fetchall()
            field_names = [desc[0] for desc in self.cur.description]
            df = pd.DataFrame(rows, columns=field_names)
            return df
        except Exception as e:
            self.logger.error(f'【posgresql-fetch_data】异常, {query} - {str(e)}')
        return pd.DataFrame()


def main(query='select * from footprint_candle limit 5;'):
        
    # alias
    host, port, user, passwd, database, charset  = yml_utils.get_posgresql_info(os.getcwd())
    kaqBtcTimeScaleDbRepository = KaqQuantPostgreSqlRepository(host, port, user, passwd, database)
    df = kaqBtcTimeScaleDbRepository.fetch_data(query)
    print(df.head(3))

if __name__ == '__main__':
    main()
